#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""TODO""" # TODO

__all__ = ["SurrogateAssistedAlg", "SAIL"]


########### IMPORTS ########### {{{1

import abc
from timeit import default_timer as timer
from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, Mapping, MutableMapping, overload
import warnings
import numpy as np
import random
import copy
from functools import partial
from sklearn.gaussian_process import GaussianProcessRegressor


from .base import *
from .evolution import *
from qdpy.utils import *
from qdpy.base import *
from qdpy.containers import *
from qdpy import tools


########### SURROGATE-ASSISTED ALGORITHMS ########### {{{1


class SurrogateAssistedAlg(AlgWithBatchSuggestions):
    """TODO"""

    _illumination_algo: QDAlgorithm
    _acquisition_algo: QDAlgorithm
    _prediction_algo: Optional[QDAlgorithm]
    max_selected_acquisition_suggestions: float # float instead of int to allow `np.inf`
    budget_not_suggested_by_surrogate: float # float instead of int to allow `np.inf`


    def __init__(self,
            illumination_algo: QDAlgorithm,
            acquisition_algo: QDAlgorithm,
            prediction_algo: Optional[QDAlgorithm] = None,
            max_selected_acquisition_suggestions: float = np.inf,
            budget_not_suggested_by_surrogate: float = np.inf,
            **kwargs):
        self._illumination_algo = illumination_algo
        self._acquisition_algo = acquisition_algo
        self._prediction_algo = prediction_algo if prediction_algo else copy.deepcopy(self._acquisition_algo)
        self.max_selected_acquisition_suggestions = max_selected_acquisition_suggestions
        self.budget_not_suggested_by_surrogate = budget_not_suggested_by_surrogate

        kwargs2 = copy.deepcopy(kwargs)
        def del_if_exists(keys: Sequence[str]) -> None:
            for k in keys:
                if k in kwargs2:
                    del kwargs2[k]
        del_if_exists(["container", "budget", "dimension", "batch_size", "nb_objectives", "optimisation_task", "base_ind_gen"])

        super().__init__(illumination_algo.container, illumination_algo.budget,
                dimension=illumination_algo.dimension, batch_size=illumination_algo.batch_size,
                nb_objectives=illumination_algo._nb_objectives, optimisation_task=illumination_algo.optimisation_task,
                base_ind_gen=illumination_algo._base_ind_gen, **kwargs2)


    @property
    def illumination_algo(self) -> QDAlgorithm:
        return self._illumination_algo

    @property
    def acquisition_algo(self) -> QDAlgorithm:
        return self._acquisition_algo

    @property
    def prediction_algo(self) -> Optional[QDAlgorithm]:
        return self._prediction_algo


    @abc.abstractmethod
    def _compute_acquisition(self) -> List[IndividualLike]:
        """Compile and return acquisition suggestions."""
        raise NotImplementedError("`self._compute_acquisition` must be implemented by sub-classes")


    def _select_suggestions_from_acquisition(self, acquisition_suggestions: List[IndividualLike],
            nb_suggestions: int) -> List[IndividualLike]:
        """Select `nb_suggestions` from `acquisition_suggestions` to be used as suggestions for `self._illumination_algo`.
        By default, select randomly (uniformly) `self.max_selected_acquisition_suggestions` individuals among `acquisition_suggestions`.

        Parameters
        ----------
        :param acquisition_suggestions: Sequence[IndividualLike]
            The sequence of proposed suggestions from acquisition.
        :param nb_suggestions: int
            The maximal number of suggestions to include in the resulting sequence of suggestions.
        """
        nb = int(min(nb_suggestions, self.max_selected_acquisition_suggestions))
        if len(acquisition_suggestions) <= nb:
            return acquisition_suggestions
        else:
            random.shuffle(acquisition_suggestions)
            return acquisition_suggestions[:nb]


    def _generate_batch_suggestions(self) -> None:
        """Compute all suggestions for the current batch and store them in `self._batch_inds`."""
        # Compute acquisition if the container is not empty
        acquisition_suggestions: List[IndividualLike] = self._compute_acquisition() if len(self.container) > 0 else []
        # Select suggestions from the acquisition container
        self._batch_inds.extend(self._select_suggestions_from_acquisition(acquisition_suggestions, self.batch_size))
        # If not enough suggestions were obtained from acquisition, ask `self._illumination_algo` for suggestions.
        nb_not_suggested_by_surrogate = int(min(self.budget_not_suggested_by_surrogate, self.batch_size - len(self._batch_inds)))
        for _ in range(nb_not_suggested_by_surrogate):
            self._batch_inds.append(self._illumination_algo.ask())


    def _internal_tell(self, individual: IndividualLike, added_to_container: bool) -> None:
        self._illumination_algo._tell_after_container_update(individual, added_to_container)



@registry.register
class SAIL(SurrogateAssistedAlg):
    """TODO"""

    ucb_stddev: float
    _gp: Any

    def __init__(self, ucb_stddev: float = 1.0, **kwargs):
        super().__init__(**kwargs)
        self.ucb_stddev = ucb_stddev
        self._gp = None

    def __getstate__(self):
        odict = super().__getstate__()
        del odict['_gp']
        return odict


    def _acquisition_eval_fn(self, inds: Sequence[IndividualLike]) -> Sequence[Tuple[FitnessValuesLike, FeaturesLike]]:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            y_mean, y_std = self._gp.predict(np.array(inds), return_std = True)
        ucb_val = y_mean + self.ucb_stddev * y_std.reshape((-1, 1))
        res = []
        for i in range(len(inds)):
            fitness = []
            for j in range(self._nb_objectives):
                fitness.append(ucb_val[i, j])
            features = []
            for j in range(ucb_val.shape[1] - self._nb_objectives):
                features.append(ucb_val[i, j + self._nb_objectives])
            res.append((fitness, features))
        return res


    def _create_gaussian_regressor(self) -> None:
        # Create dataset
        dataX = []
        dataY = []
        for ind in self.container:
            dataX.append(ind)
            dataY.append(list(ind.fitness.values) + list(ind.features))
        # Fit gaussian process on the dataset
        self._gp = GaussianProcessRegressor()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._gp.fit(dataX, dataY)


    def _compute_acquisition(self) -> List[IndividualLike]:
        """Compile and return acquisition suggestions."""
        # Reinit `self._acquisition_algo`
        self._acquisition_algo.container.clear()
        # TODO reinit ??

        # Create gaussian regressor
        self._create_gaussian_regressor()
        # Run `self._acquisition_algo` optimisation process
        try:
            self._acquisition_algo.optimise(self._acquisition_eval_fn, executor=None, send_several_suggestions_to_fn=True)
        except RuntimeError:
            pass
        # Retrieve new suggestions
        suggestions = copy.deepcopy(list(set(self._acquisition_algo.container) - set(self.container)))
        for ind in suggestions:
            del ind.fitness.values
            ind.features = ()
        #print("DEBUG _compute_acquisition", len(suggestions))
        return suggestions


    # TODO sobol
    def _select_suggestions_from_acquisition(self, acquisition_suggestions: List[IndividualLike],
            nb_suggestions: int) -> List[IndividualLike]:
        """Select `nb_suggestions` from `acquisition_suggestions` to be used as suggestions for `self._illumination_algo`.
        By default, select randomly (uniformly) `self.max_selected_acquisition_suggestions` individuals among `acquisition_suggestions`.

        Parameters
        ----------
        :param acquisition_suggestions: Sequence[IndividualLike]
            The sequence of proposed suggestions from acquisition.
        :param nb_suggestions: int
            The maximal number of suggestions to include in the resulting sequence of suggestions.
        """
        nb = int(min(nb_suggestions, self.max_selected_acquisition_suggestions))
        if len(acquisition_suggestions) <= nb:
            return acquisition_suggestions
        else:
            random.shuffle(acquisition_suggestions)
            return acquisition_suggestions[:nb]




# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""The :mod:`~qdpy.utils` module is a collection of small functions and classes handling common patterns."""

#__all__ = ["jit"]

from collections.abc import Iterable
from typing import Optional, Tuple, TypeVar, Union, Any, MutableSet, Mapping, MutableMapping, Sequence, MutableSequence, Callable, Tuple
from typing_extensions import runtime, Protocol
import inspect


########### UTILS ########### {{{1

def is_iterable(obj: Any) -> bool:
    """Return if ``obj`` is iterable or not."""
    return isinstance(obj, Iterable)

#def in_bounds(val: Union[float, Sequence[float]], domain: Union[DomainLike, Sequence[DomainLike]]) -> bool:
def in_bounds(val: Any, domain: Any) -> bool:
    """Return if ``val`` (if a value) or all values in ``val`` (if an iterable) are in ``domain``."""
    #assert len(domain) >= 2, f"``domain`` must be a 2-tuple of numbers or a sequence of 2-tuples of numbers."
    if isinstance(val, Sequence):
        if isinstance(domain[0], Sequence):
            if len(val) == len(domain):
                return all((v >= d[0] and v <= d[1] for v, d in zip(val, domain)))
            else:
                raise ValueError(f"if ``val`` is a Sequence, ``domain`` must have the same length as ``val``.")
        else:
            return all((v >= domain[0] and v <= domain[1] for v in val))
    else:
        if isinstance(domain[0], Sequence):
            raise ValueError(f"if ``val`` is not a Sequence, ``domain`` must be a 2-tuple of numbers.")
        else:
            return val >= domain[0] and val <= domain[1]

#def _hashify(item):
#    """Verify if *item* is hashable, if not, try and return it as a tuple."""
#    if isinstance(item, collections.abc.Hashable):
#        return item
#    else:
#        return tuple(item)

def tuplify(item: Union[Any, Sequence[Any]]) -> Tuple[Any, ...]:
    if isinstance(item, Sequence):
        return tuple(item)
    else:
        return (item,)




########### NUMBA ########### {{{1
def _dummyJit(*args, **kwargs):
    """
    Dummy version of jit decorator, does nothing
    """
    if len(args) == 1 and callable(args[0]):
        return args[0]
    else:
        def wrap(func):
            return func
        return wrap
try:
    import numba
    from numba import jit
except ImportError:
    jit = _dummyJit



# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""TODO"""
#from __future__ import annotations

__all__ = ["OrderedSet", "Container", "Grid", "CVTGrid", "NoveltyArchive"]

########### IMPORTS ########### {{{1
import sys
import warnings
import abc
import collections.abc
import math
from functools import reduce
import operator
import numpy as np
from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, MutableMapping, Mapping, overload
from typing_extensions import runtime, Protocol
import traceback

from qdpy.utils import *
from qdpy.base import *
from qdpy.metrics import novelty, novelty_local_competition, novelty_nn



########### BACKEND CLASSES ########### {{{1

@registry.register
class OrderedSet(MutableSet[T], Sequence[T]):
    """A MutableSet variant that conserves entries order, and can be accessed like a Sequence.
    This implementation is not optimised, but does not requires the type ``T`` of items to be hashable.
    It also does not implement indexing by slices.
    
    Parameters
    ----------
    iterable: Optional[Iterable[T]]
        items to add to the OrderedSet
    """

    _items: List[T]    # Internal storage

    def __init__(self, iterable: Optional[Iterable] = None) -> None:
        self._items = []
        if iterable is not None:
            self.update(iterable)

    def __len__(self) -> int:
        return len(self._items)

    @overload
    def __getitem__(self, index: int) -> T: ...

    @overload
    def __getitem__(self, s: slice) -> Sequence[T]: ...

    def __getitem__(self, key):
        if isinstance(key, slice):
            raise NotImplementedError
        return self._items[key]

    def __contains__(self, key: Any) -> bool:
        return key in self._items

    def __iter__(self) -> Iterator[T]:
        return iter(self._items)

    def __reversed__(self) -> Iterator[T]:
        return reversed(self._items)

    def __repr__(self) -> str:
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self))

    def __delitem__(self, idx) -> None:
        del self._items[idx]

    def count(self, key: T) -> int:
        return 1 if key in self else 0

    def index(self, key: T, start: int = 0, stop: int = sys.maxsize) -> int:
        return self._items.index(key, start, stop)

    def add(self, key: T) -> None:
        """Add ``key`` to this OrderedSet, if it is not already present. """
        try:
            self._items.index(key)
        except ValueError:
            self._items.append(key)

    def discard(self, key: T) -> None:
        """Discard ``key`` in this OrderedSet. Does not raise an exception if absent."""
        try:
            self._items.remove(key)
        except ValueError:
            return

    def update(self, iterable: Iterable) -> None:
        """Add the items in ``iterable``, if they are not already present in the OrderedSet.  """
        try:
            for item in iterable:
                self.add(item)
        except TypeError:
            raise ValueError(f"Argument needs to be an iterable, got {type(iterable)}")


BackendLike = Union[MutableSequence[T], OrderedSet[T]]

########### CONTAINER CLASSES ########### {{{1

# TODO verify that containers are thread-safe
@registry.register
class Container(Sequence, Summarisable, Copyable, CreatableFromConfig):
    """TODO
    
    Parameters
    ----------
    iterable: Iterable[IndividualLike] or None
        TODO
    storage_type: Backend (MutableSet or MutableSequence)
        TODO
    depot_type: bool or Backend (MutableSet or MutableSequence)
        TODO
    fitness_domain: Sequence[DomainLike] (sequence of 2-tuple of numbers)
        TODO
    features_domain: Sequence[DomainLike] (sequence of 2-tuple of numbers)
        TODO
    """ # TODO

    name: Optional[str]
    items: BackendLike[IndividualLike]
    depot: Optional[BackendLike[IndividualLike]]
    fitness_domain: Optional[Sequence[DomainLike]]
    features_domain: Optional[Sequence[DomainLike]]
    recentness: MutableSequence[int]
    _capacity: float
    _size: int
    _best: Optional[IndividualLike]
    _best_fitness: Optional[FitnessLike]
    _best_features: Optional[FeaturesLike]
    _nb_discarded: int
    _nb_added: int
    _nb_rejected: int

    def __init__(self, iterable: Optional[Iterable] = None,
            storage_type: Type[BackendLike] = list, depot_type: Union[bool, Type[BackendLike]] = False,
            fitness_domain: Optional[Sequence[DomainLike]] = None,
            features_domain: Optional[Sequence[DomainLike]] = None,
            capacity: Optional[float] = None, name: Optional[str] = None,
            **kwargs: Any) -> None:
        self.items = storage_type()
        if depot_type is True:
            self.depot = storage_type()
        elif not isinstance(depot_type, bool):
            self.depot = depot_type()
        else:
            self.depot = None
        self.fitness_domain = fitness_domain
        if self.fitness_domain is not None:
            for f in self.fitness_domain:
                if not is_iterable(f) or len(f) != 2:
                    raise ValueError("``fitness_domain`` must be a sequence of 2-tuples.")
        self.features_domain = features_domain
        if self.features_domain is not None:
            for f in self.features_domain:
                if not is_iterable(f) or len(f) != 2:
                    raise ValueError("``features_domain`` must be a sequence of 2-tuples.")
        self.recentness = []
        self.name = name if name is not None else f"{self.__class__.__name__}-{id(self)}"
        self._capacity = math.inf if capacity is None else capacity
        self._size = 0
        self._best = None
        self._best_fitness = None
        self._best_features = None
        self._nb_discarded = 0
        self._nb_rejected = 0
        self._nb_added = 0
        if iterable is not None:
            self.update(iterable)

    @property
    def capacity(self) -> float:
        """Return the capacity of the container (i.e. maximal number of items/spots/bins/etc). Can be math.inf."""
        return self._capacity

    @property
    def free(self) -> float:
        """Return the number of free spots in the container. Can be math.inf."""
        return self._capacity - self._size

    @property
    def size(self) -> int:
        """Return the size of the container (i.e. number of items, spots, bins, etc)."""
        return self._size

    @property
    def best(self) -> Any:
        """Return the best individual. """
        return self._best

    @property
    def best_fitness(self) -> Optional[FitnessLike]:
        """Return the fitness values of the individual with the best quality, or None. """
        return self._best_fitness

    @property
    def best_features(self) -> Optional[FeaturesLike]:
        """Return the features values of the individual with the best quality, or None. """
        return self._best_features

    @property
    def nb_discarded(self) -> int:
        """Return the number of individuals discarded by the container since its creation. """
        return self._nb_discarded

    @property
    def nb_added(self) -> int:
        """Return the number of individuals added into the container since its creation. """
        return self._nb_added

    @property
    def nb_rejected(self) -> int:
        """Return the number of individuals rejected (when added) by the container since its creation. """
        return self._nb_rejected

    @property
    def nb_operations(self) -> int:
        """Return the number of adds, modifications and discards since the creation of this container. """
        return self._nb_added + self._nb_discarded


    @property
    def fitness_extrema(self) -> Optional[Sequence[DomainLike]]:
        """Return the extrema values of the fitnesses of the stored individuals."""
        if len(self) == 0:
            return None
        maxima = np.array(self[0].fitness.values) # type: ignore
        minima = np.array(self[0].fitness.values) # type: ignore
        for ind in self:
            ind_f = ind.fitness.values
            for i in range(len(maxima)):
                if ind_f[i] > maxima[i]:
                    maxima[i] = ind_f[i]
                elif ind_f[i] < minima[i]:
                    minima[i] = ind_f[i]
        return tuple(zip(minima, maxima))


    @property
    def features_extrema(self) -> Optional[Sequence[DomainLike]]:
        """Return the extrema values of the features of the stored individuals."""
        if len(self) == 0:
            return None
        maxima = np.array(self[0].features) # type: ignore
        minima = np.array(self[0].features) # type: ignore
        for ind in self:
            ind_f = ind.features
            for i in range(len(maxima)):
                if ind_f[i] > maxima[i]:
                    maxima[i] = ind_f[i]
                elif ind_f[i] < minima[i]:
                    minima[i] = ind_f[i]
        return tuple(zip(minima, maxima))



    def size_str(self) -> str:
        """Return a string describing the fullness of the container."""
        if math.isinf(self.capacity):
            return str(self.size)
        else:
            return "%i/%i" % (self.size, self.capacity)

    def __len__(self) -> int:
        return len(self.items)

    @overload
    def __getitem__(self, index: int) -> T: ...

    @overload
    def __getitem__(self, s: slice) -> Sequence[IndividualLike]: ...

    def __getitem__(self, key):
        if isinstance(key, slice):
            raise NotImplementedError
        return self.items[key]

    def __contains__(self, key: Any) -> bool:
        return key in self.items

    def __iter__(self) -> Iterator[IndividualLike]:
        return iter(self.items)

#    def __reversed__(self) -> Iterator[IndividualLike]:
#        return reversed(self.items)

    def __repr__(self) -> str:
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self))


    def _add_to_collection(self, collection: BackendLike[IndividualLike], individual: IndividualLike) -> Tuple[bool,int]:
        """Add ``individual`` to ``collection``.
        Return a tuple containing (added, index), with ``added`` a bool saying whether ``individual`` was added or not to ``collection``, and ``index`` the index in the ``collection``.
        ``collection`` can be a MutableSequence or an ordered set implementing Sequence and MutableSet."""
        old_len: int = len(collection)
        if isinstance(collection, MutableSet):
            collection.add(individual)
            if len(collection) == old_len:
                return False, collection.index(individual)
            else:
                return True, len(collection) - 1
        elif isinstance(collection, MutableSequence):
            added: bool = False
            try:
                index: int = collection.index(individual)
            except ValueError:
                collection.append(individual)
                index = len(collection) - 1
                added = True
            return added, index
        else:
            raise ValueError("collection must be an ordered set implementing MutableSet or a Sequence")


    def in_bounds(self, val: Any, domain: Any) -> bool:
        """TODO"""
        if domain is None or len(domain) == 0:
            return True
        return in_bounds(val, domain)


    def _check_if_can_be_added(self, individual: IndividualLike) -> None:
        """TODO"""
        # Retrieve features and fitness from individual
        #ind_fitness: FitnessValuesLike = self._get_fitness_from_ind(individual)
        #ind_features: FeaturesLike = self._get_features_from_ind(individual)
        exc = None
        if not individual.fitness.valid:
            exc = ValueError(f"Fitness is not valid.")
        # Check if fitness and features are out of bounds
        if not self.in_bounds(individual.fitness.values, self.fitness_domain):
            exc = ValueError(f"fitness ({str(individual.fitness.values)}) out of bounds ({str(self.fitness_domain)}).")
        if not self.in_bounds(individual.features, self.features_domain):
            exc = ValueError(f"features ({str(individual.features)}) out of bounds ({str(self.features_domain)}).")
        if exc:
            self._nb_rejected += 1
            raise exc

    #def _add_internal(self, individual: T, raise_if_not_added_to_depot: bool, only_to_depot: bool, ind_fitness: FitnessLike, ind_features: FeaturesLike) -> Optional[int]:
    def _add_internal(self, individual: IndividualLike, raise_if_not_added_to_depot: bool, only_to_depot: bool) -> Optional[int]:
        """TODO"""
        # Verify if we do not exceed capacity
        if self.free < 1:
            self._nb_rejected += 1
            raise IndexError(f"No free slot available in this container.")
        if raise_if_not_added_to_depot and self.depot is None:
            self._nb_rejected += 1
            raise ValueError(f"`raise_if_not_added_to_depot` can only be set to True if a depot exists.")
        if only_to_depot and self.depot is None:
            self._nb_rejected += 1
            raise ValueError(f"`only_to_depot` can only be set to True if a depot exists.")
        # Add to depot, if needed
        added_depot, index_depot = self._add_to_collection(self.depot, individual) if self.depot is not None else False, 0
        if raise_if_not_added_to_depot and not added_depot:
            self._nb_rejected += 1
            raise ValueError(f"Individual could not be added to the depot.")
        if only_to_depot:
            return None
        else:
            # Add to storage
            added, index = self._add_to_collection(self.items, individual)
            # Update best_fitness
            if added:
                #if self._best is None or self._dominates(individual, self._best):
                if self._best is None or individual.fitness.dominates(self._best_fitness):
                    self._best = individual
                    self._best_fitness = individual.fitness
                    self._best_features = individual.features
                self.recentness.append(self._nb_added)
                self._size += 1
                self._nb_added += 1
            return index

    def add(self, individual: IndividualLike, raise_if_not_added_to_depot: bool = False) -> Optional[int]:
        """Add ``individual`` to the container, and returns its index, if successful, None elsewise. If ``raise_if_not_added_to_depot`` is True, it will raise and exception if it was not possible to add it also to the depot."""
        # Retrieve features and fitness from individual and check if they are not out-of-bounds
        self._check_if_can_be_added(individual)
        # Add
        return self._add_internal(individual, raise_if_not_added_to_depot, False)

    def _discard_by_index(self, individual: IndividualLike, idx: Optional[int] = None, idx_depot: Optional[int] = None, also_from_depot: bool = False) -> None:
        """Remove ``individual`` of the container. If ``also_from_depot`` is True, discard it also from the depot, if it exists. Use the indexes ``idx`` and ``idx_depot`` if they are provided."""
        # Remove from depot
        if also_from_depot and self.depot is not None:
            if idx_depot is None:
                try:
                    self.depot.remove(individual)
                except KeyError:
                    pass
            else:
                del self.depot[idx_depot]
        # Remove from container
        if idx is None:
            try:
                idx = self.items.index(individual)
            except KeyError:
                return
        del self.items[idx]
        del self.recentness[idx]
        self._size -= 1
        self._nb_discarded += 1
        if self._size < 0:
            raise RuntimeError("`self.size` < 0 !")


    def discard(self, individual: IndividualLike, also_from_depot: bool = False) -> None:
        """Remove ``individual`` of the container. If ``also_from_depot`` is True, discard it also from the depot, if it exists."""
        idx_depot = None
        idx = None
        if also_from_depot and self.depot is not None:
            try:
                idx_depot = self.depot.index(individual)
            except KeyError:
                pass
        # Remove from container
        try:
            idx = self.items.index(individual)
        except KeyError:
            pass
        self._discard_by_index(individual, idx, idx_depot, also_from_depot)


    def update(self, iterable: Iterable, ignore_exceptions: bool = True, issue_warning: bool = False) -> int:
        """Add the individuals in ``iterable``, if they are not already present in the container.
        If ``ignore_exceptions`` is True, it will ignore exceptions raised when adding each individual, but issue warnings instead.
        Returns the number of elements inserted or updated."""
        nb_inserted: int = 0
        item_index: Optional[int] = None
        try:
            if ignore_exceptions and issue_warning:
                for item in iterable:
                    try:
                        item_index = self.add(item)
                        if item_index is not None:
                            nb_inserted += 1
                    except IndexError as e:
                        warnings.warn(f"Adding individual failed (index out of bounds): {str(e)}")
                    except ValueError as e:
                        warnings.warn(f"Adding individual failed (attribute out of bounds): {str(e)}")
                    except Exception as e:
                        warnings.warn(f"Adding individual failed: {str(e)}")
                        traceback.print_exc()
            elif ignore_exceptions:
                for item in iterable:
                    try:
                        item_index = self.add(item)
                        if item_index is not None:
                            nb_inserted += 1
                    except Exception:
                        pass
            else:
                for item in iterable:
                    item_index = self.add(item)
                    if item_index is not None:
                        nb_inserted += 1
        except TypeError:
            raise ValueError(f"Argument needs to be an Iterable, got {type(Iterable)}")
        return nb_inserted


    def clear(self) -> None:
        """Clear all individual in the collection (but not those in the depot, see method ``clear_all``)."""
        items = list(self)
        for e in items:
            self.discard(e)

    def clear_all(self) -> None:
        """Clear all individual in the collection AND in the depot."""
        self.clear()
        if self.depot is not None:
            self.depot.clear()


    def novelty(self, individual: IndividualLike, **kwargs):
        """Returns the novelty score of `individual`, using the depot as archive. TODO""" # TODO
        if self.depot is None:
            raise RuntimeError(f"A depot is necessary to assess novelty.")
        return novelty(individual, self.depot, **kwargs)

    def novelty_local_competition(self, individual: IndividualLike, **kwargs):
        """Returns the novelty and local competition scores of `individual`, using the depot as archive. TODO""" # TODO
        if self.depot is None:
            raise RuntimeError(f"A depot is necessary to assess novelty.")
        return novelty_local_competition(individual, self.depot, **kwargs)


    def to_grid(self, shape: Union[ShapeLike, int],
            max_items_per_bin: int = 1,
            capacity: Optional[float] = None, **kwargs: Any) -> Any:
        """Return a grid representation of this container, with `shape` the shape of the grid.
        
        Parameters
        ----------
        :param shape: Union[ShapeLike, int]
            The shape of the grid.
        :param max_items_per_bin: int
            The maximal number of entries stored in each bin of the grid. Defaults to 1.
        :param fitness_domain: Optional[Sequence[DomainLike]]
            The domain of the fitness of the individual of this grid. Default to `self.fitness_extrema`.
        :param features_domain: Optional[Sequence[DomainLike]]
            The domain of the features of the individual of this grid. Default to `self.features_extrema`.
        :param capacity: Optional[float] = None
            The capacity (i.e. maximal number of entries) of the returned grid representation. Default to None (i.e. no limit).
        :param storage_type: Type[BackendLike]
            How individuals are stored internally. Defaults to list.

        Return
        ------
        grid: Grid
            Grid representation of this container.
        """
        if not 'fitness_domain' in kwargs:
            kwargs['fitness_domain'] = self.fitness_domain if self.fitness_domain is not None else self.fitness_extrema
        if not 'features_domain' in kwargs:
            kwargs['features_domain'] = self.features_domain if self.features_domain is not None else self.features_extrema
        return Grid(self, shape=shape, max_items_per_bin=max_items_per_bin, capacity=capacity, **kwargs)





# TODO Container is already a population ?
@registry.register
class Population(Container):
    """TODO""" # TODO
    pass



#def _nparray_from_dict(d: dict, shape: ShapeLike, default_val: float = np.nan, dtype: Type = float) -> np.array:
#    res = np.full(shape, default_val, dtype = dtype)
#    for k,v in d:
#        res[k] = v
#    return res



########### GRID-BASED CLASSES ########### {{{2

# Custom types
GridSolutionsLike = MutableMapping[GridIndexLike, MutableSequence]
GridItemsPerBinLike = MutableMapping[GridIndexLike, int]
GridFitnessLike = MutableMapping[GridIndexLike, MutableSequence[FitnessLike]]
GridFeaturesLike = MutableMapping[GridIndexLike, MutableSequence[FeaturesLike]]
GridQualityLike = MutableMapping[GridIndexLike, Optional[FitnessLike]]
GridRecentnessPerBinLike = MutableMapping[GridIndexLike, MutableSequence[int]]


@registry.register
class Grid(Container):
    """TODO""" # TODO

    fitness_domain: Sequence[DomainLike]
    features_domain: Sequence[DomainLike]

    _shape: ShapeLike
    _max_items_per_bin: int
    _filled_bins: int
    _solutions: GridSolutionsLike
    _nb_items_per_bin: GridItemsPerBinLike
    _fitness: GridFitnessLike
    _features: GridFeaturesLike
    _quality: GridQualityLike
    _quality_array: np.array
    _bins_size: Sequence[float]
    _nb_bins: int
    recentness_per_bin: GridRecentnessPerBinLike
    history_recentness_per_bin: GridRecentnessPerBinLike
    activity_per_bin: np.array


    def __init__(self, iterable: Optional[Iterable] = None,
            shape: Union[ShapeLike, int] = (1,), max_items_per_bin: int = 1,
            fitness_domain: Optional[Sequence[DomainLike]] = ((0., np.inf),),
            **kwargs: Any) -> None:
        self._shape = tuplify(shape)
        self._max_items_per_bin = max_items_per_bin
        super().__init__([], fitness_domain=fitness_domain, **kwargs)
        if self.features_domain is None or len(self.features_domain) == 0:
            raise ValueError("`features_domain` must be specified and have a length > 0.")
        if self.fitness_domain is None or len(self.fitness_domain) == 0:
            raise ValueError("`fitness_domain` must be specified and have a length > 0.")
        #if len(self.features_domain) != len(self.shape):
        #    raise ValueError("`features_domain` must have the same shape as `shape`.")
        self._init_grid()
        if not "capacity" in kwargs:
            self._capacity = self._nb_bins * self._max_items_per_bin
        if iterable is not None:
            self.update(iterable)


    def _init_grid(self) -> None:
        """Initialise the grid to correspond to the shape `self.shape`."""
        self._solutions = {x: [] for x in self._index_grid_iterator()}
        self._nb_items_per_bin = np.zeros(self._shape, dtype=float) #{x: 0 for x in self._index_grid_iterator()}
        self._fitness = {x: [] for x in self._index_grid_iterator()}
        self._features = {x: [] for x in self._index_grid_iterator()}
        self._quality = {x: None for x in self._index_grid_iterator()}
        self._quality_array = np.full(self._shape + (len(self.fitness_domain),), np.nan)
        self._bins_size = [(self.features_domain[i][1] - self.features_domain[i][0]) / float(self.shape[i]) for i in range(len(self.shape))]
        self._filled_bins = 0
        self._nb_bins = reduce(operator.mul, self._shape)
        self.recentness_per_bin = {x: [] for x in self._index_grid_iterator()}
        self.history_recentness_per_bin = {x: [] for x in self._index_grid_iterator()}
        self.activity_per_bin = np.zeros(self._shape, dtype=float)


    @property
    def shape(self) -> ShapeLike:
        """Return the shape of the grid."""
        return self._shape

    @property
    def max_items_per_bin(self) -> int:
        """Return the maximal number of items stored in a bin of the grid."""
        return self._max_items_per_bin

    @property
    def filled_bins(self) -> int:
        """Return the number of filled bins of the container."""
        return self._filled_bins

    @property
    def solutions(self) -> GridSolutionsLike:
        """Return the solutions in the grid."""
        return self._solutions

    @property
    def nb_items_per_bin(self) -> GridItemsPerBinLike:
        """Return the number of items stored in each bin of the grid."""
        return self._nb_items_per_bin

    @property
    def fitness(self) -> GridFitnessLike:
        """Return the fitness values in the grid."""
        return self._fitness

    @property
    def features(self) -> GridFeaturesLike:
        """Return the features values in the grid."""
        return self._features

    @property
    def quality(self) -> GridQualityLike:
        """Return the best fitness values in the grid."""
        return self._quality

    @property
    def quality_array(self) -> np.array:
        """Return the best fitness values in the grid, as a numpy array."""
        return self._quality_array

    @property
    def best_index(self) -> Optional[GridIndexLike]:
        """Return the index of the individual with the best quality, or None. """
        if self._best_features is None:
            return None
        else:
            return self.index_grid(self._best_features)


    def filled_str(self) -> str:
        """Return a string describing the fullness of the grid (not the container itself, which is handled by ``Container.size_str``)."""
        return "%i/%i" % (self.filled_bins, self._nb_bins)


    def index_grid(self, features: FeaturesLike) -> GridIndexLike:
        """Get the index in the grid of a given individual with features ``features``, raising an IndexError if it is outside the grid. """
        index: List[int] = []
        if len(features) != len(self.shape):
            raise IndexError(f"Length of parameter ``features`` ({len(features)}) does not corresponds to the number of dimensions of the grid ({len(self.shape)}).")
        for i in range(len(features)):
            normalised_feature: float = features[i] - self.features_domain[i][0]
            if normalised_feature == self.features_domain[i][1] - self.features_domain[i][0]:
                partial: int = self.shape[i] - 1
            elif normalised_feature > self.features_domain[i][1] - self.features_domain[i][0]:
                raise IndexError(f"``features`` ({str(features)}) out of bounds ({str(self.features_domain)})")
            else:
                partial = int(normalised_feature / self._bins_size[i])
            index.append(partial)
        return tuple(index)


    def _index_grid_iterator(self) -> Generator[GridIndexLike, None, None]:
        """Return an iterator of the index of the grid, based on ``self.shape``."""
        val: List[int] = [0] * len(self._shape)
        yield tuple(val)
        while True:
            for i in reversed(range(len(self._shape))):
                val[i] += 1
                if val[i] >= self._shape[i]:
                    if i == 0:
                        return
                    val[i] = 0
                else:
                    yield tuple(val)
                    break

    def _update_quality(self, ig: GridIndexLike) -> None:
        """Update quality in bin ``ig`` of the grid."""
        if self._nb_items_per_bin[ig] == 0:
            val: Optional[FitnessLike] = None
        elif self._nb_items_per_bin[ig] == 1:
            val = self.fitness[ig][0]
        else:
            best: IndividualLike = self.solutions[ig][0]
            for s in self.solutions[ig][1:]:
                if s.fitness.dominates(best.fitness):
                    best = s
            val = best.fitness
        self.quality[ig] = val
        if val is None:
            self.quality_array[ig] = math.nan
        else:
            self.quality_array[ig] = val.values


    def add(self, individual: IndividualLike, raise_if_not_added_to_depot: bool = False) -> Optional[int]:
        """Add ``individual`` to the grid, and returns its index, if successful, None elsewise. If ``raise_if_not_added_to_depot`` is True, it will raise and exception if it was not possible to add it also to the depot."""
        # Retrieve features and fitness from individual and check if they are not out-of-bounds
        self._check_if_can_be_added(individual)
        # Find corresponding index in the grid
        ig = self.index_grid(individual.features) # Raise exception if features are out of bounds

        # Check if individual can be added in grid, if there are enough empty spots
        can_be_added: bool = False
        if self._nb_items_per_bin[ig] < self.max_items_per_bin:
            can_be_added = True
        else:
            worst_idx = 0
            worst: IndividualLike = self.solutions[ig][worst_idx]
            if self._nb_items_per_bin[ig] > 1:
                for i in range(1, self._nb_items_per_bin[ig]):
                    s = self.solutions[ig][i]
                    if worst.fitness.dominates(s.fitness):
                        worst = s
                        worst_idx = i
            if individual.fitness.dominates(worst.fitness):
                Container.discard(self, self.solutions[ig][worst_idx])
                self._discard_from_grid(ig, worst_idx)
                can_be_added = True

        # Add individual in grid, if there are enough empty spots
        if can_be_added:
            if self._nb_items_per_bin[ig] == 0:
                self._filled_bins += 1
            # Add individual in container
            old_len: int = self._size
            index: Optional[int] = self._add_internal(individual, raise_if_not_added_to_depot, False)
            if index == old_len: # Individual was not already present in container
                self._solutions[ig].append(individual)
                self._fitness[ig].append(individual.fitness)
                self._features[ig].append(individual.features)
                self.recentness_per_bin[ig].append(self._nb_added)
                self.history_recentness_per_bin[ig].append(self._nb_added)
                self._nb_items_per_bin[ig] += 1
                self.activity_per_bin[ig] += 1
            # Update quality
            self._update_quality(ig)
            return index
        else:
            # Only add to depot
            if self.depot is not None:
                self._add_internal(individual, raise_if_not_added_to_depot, True)
            return None

    def _discard_from_grid(self, ig: GridIndexLike, index_in_bin: int) -> None:
        # Remove individual from grid
        del self._solutions[ig][index_in_bin]
        self._nb_items_per_bin[ig] -= 1
        del self._fitness[ig][index_in_bin]
        del self._features[ig][index_in_bin]
        del self.recentness_per_bin[ig][index_in_bin]
        # Update quality
        self._update_quality(ig)
        # Update number of filled bins
        if self._nb_items_per_bin[ig] == 0:
            self._filled_bins -= 1


    def discard(self, individual: IndividualLike, also_from_depot: bool = False) -> None:
        """Remove ``individual`` of the container. If ``also_from_depot`` is True, discard it also from the depot, if it exists."""
        old_len: int = self._size
        # Remove individual from container
        Container.discard(self, individual, also_from_depot)
        if self._size == old_len:
            return
        # Remove individual from grid
        ig = self.index_grid(individual.features) # Raise exception if features are out of bounds
        index_in_bin = self.solutions[ig].index(individual)
        self._discard_from_grid(ig, index_in_bin)




from sklearn.cluster import KMeans
from sklearn.gaussian_process import GaussianProcessRegressor

@registry.register
class CVTGrid(Grid):
    """TODO""" # TODO

    _grid_shape: ShapeLike
    _nb_sampled_points: int
    cluster_centers: np.array

    def __init__(self, iterable: Optional[Iterable] = None,
            shape: Union[ShapeLike, int] = (1,), max_items_per_bin: int = 1,
            grid_shape: Union[ShapeLike, int] = (1,), nb_sampled_points: int = 50000, **kwargs: Any) -> None:
        self._grid_shape = tuplify(grid_shape)
        self._nb_sampled_points = nb_sampled_points
        super().__init__(iterable, shape, max_items_per_bin, **kwargs)
        if len(self.shape) != 1:
            raise ValueError("Using CVTGrid, `shape` must be a scalar or a sequence of length 1.")
        if nb_sampled_points <= 0:
            raise ValueError("`nb_sampled_points` must be positive and greatly superior to `shape` and `grid_shape`.")
        self._init_clusters()

    def _init_clusters(self) -> None:
        """Initialise the clusters and tessellate the grid."""
        sample = np.random.uniform(0.0, 1.0, (self.nb_sampled_points, len(self.grid_shape)))
        for i, d in enumerate(self.features_domain):
            sample[:, i] = d[0] + (d[1] - d[0]) * sample[:, i]
        kmeans = KMeans(init="k-means++", n_clusters=self.shape[0], n_init=1, n_jobs=1, verbose=0)
        kmeans.fit(sample)
        self.cluster_centers = kmeans.cluster_centers_

    @property
    def grid_shape(self) -> ShapeLike:
        """Return the shape of the grid."""
        return self._grid_shape

    @property
    def nb_sampled_points(self) -> int:
        """Return the number of sampled points to identify cluster centers."""
        return self._nb_sampled_points

    def index_grid(self, features: FeaturesLike) -> GridIndexLike:
        """Get the index in the cvt of a given individual with features ``features``, raising an IndexError if it is outside the cvt. """
        dists = np.empty(self._shape[0])
        for i in range(len(dists)):
            dists[i] = math.sqrt(np.sum(np.square(self.cluster_centers[i] - features)))
        return (np.argmin(dists),)



########### ARCHIVE-BASED CLASSES ########### {{{2

@registry.register
class NoveltyArchive(Container):
    """TODO""" # TODO

    depot: BackendLike[IndividualLike]

    k: int
    threshold_novelty: float
    novelty_distance: Union[str, Callable]

    def __init__(self, iterable: Optional[Iterable] = None,
            k: int = 15, threshold_novelty: float = 0.01, novelty_distance: Union[str, Callable] = "euclidean",
            depot_type: Union[bool, Type[BackendLike]] = True, **kwargs: Any) -> None:
        self.k = k
        self.threshold_novelty = threshold_novelty
        self.novelty_distance = novelty_distance
        if depot_type is None:
            raise ValueError("``depot_type`` must be specified for an archive container (either True or a BackendLike class).")
        super().__init__(iterable, depot_type=depot_type, **kwargs)


    def add(self, individual: IndividualLike, raise_if_not_added_to_depot: bool = True) -> Optional[int]:
        """Add ``individual`` to the archive, and returns its index, if successful, None elsewise. If ``raise_if_not_added_to_depot`` is True, it will raise and exception if it was not possible to add it also to the depot."""
        # Retrieve features and fitness from individual and check if they are not out-of-bounds
        self._check_if_can_be_added(individual)
        # Find novelty of this individual, and its nearest neighbour
        novelty, nn = novelty_nn(individual, self.depot, k=self.k, nn_size=1, dist=self.novelty_distance, ignore_first=False)
        if novelty > self.threshold_novelty:
            # Add individual
            return self._add_internal(individual, raise_if_not_added_to_depot, False)
        else:
            ind_nn = self.depot[nn[0]]
            if len(nn) > 0 and individual.fitness.dominates(ind_nn.fitness):
                #self.discard(nn[0])
                self._discard_by_index(ind_nn, idx_depot=nn[0])
                return self._add_internal(individual, raise_if_not_added_to_depot, False)
            else:
                return None



## MODELINE	"{{{1
## vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
## vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""Collection of test functions and benchmarks for QD algorithms based on artificial landscapes. Cf https://arxiv.org/pdf/1308.4008v1.pdf and https://arxiv.org/pdf/1207.4318.pdf and https://en.wikipedia.org/wiki/Test_functions_for_optimization and https://www.sfu.ca/~ssurjano/optimization.html"""

__all__ = [
        "sphere", "illumination_sphere", "SphereBenchmark",
        "weighted_sphere", "illumination_weighted_sphere", "WeightedSphereBenchmark",
        "rotated_hyper_ellipsoid", "illumination_rotated_hyper_ellipsoid", "RotatedHyperEllipsoidBenchmark",
        "rosenbrock", "illumination_rosenbrock", "RosenbrockBenchmark",
        "rastrigin_normalised", "illumination_rastrigin_normalised", "NormalisedRastriginBenchmark",
        "rastrigin", "illumination_rastrigin", "RastriginBenchmark",
        "schwefel", "illumination_schwefel", "SchwefelBenchmark",
        "small_schwefel", "illumination_small_schwefel", "SmallSchwefelBenchmark",
        "griewangk", "illumination_griewangk", "GriewangkBenchmark",
        "sum_of_powers", "illumination_sum_of_powers", "SumOfPowersBenchmark",
        "ackley", "illumination_ackley", "AckleyBenchmark",
        "styblinski_tang", "illumination_styblinski_tang", "StyblinskiTangBenchmark",
        "levy", "illumination_levy", "LevyBenchmark",
        "perm0db", "illumination_perm0db", "Perm0dbBenchmark",
        "permdb", "illumination_permdb", "PermdbBenchmark",
        "trid", "illumination_trid", "TridBenchmark",
        "zakharov", "illumination_zakharov", "ZakharovBenchmark",
        "dixon_price", "illumination_dixon_price", "DixonPriceBenchmark",
        "powell", "illumination_powell", "PowellBenchmark",
        "michalewicz", "illumination_michalewicz", "MichalewiczBenchmark",
        "wavy", "illumination_wavy", "WavyBenchmark",
        "trigonometric02", "illumination_trigonometric02", "Trigonometric02Benchmark",
        "qing", "illumination_qing", "QingBenchmark",
        "small_qing", "illumination_small_qing", "SmallQingBenchmark",
        "deb01", "illumination_deb01", "Deb01Benchmark",
        "shubert04", "illumination_shubert04", "Shubert04Benchmark",
        ]

from math import cos, pi
from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, Mapping, MutableMapping, overload
import numpy as np

from qdpy import algorithms
from qdpy.base import *
from .base import *


def _illuminate_artificial_landscape(ind, func, nb_features = 2):
    fitness = func(ind)
    features = list(ind[:nb_features])
    return fitness, features

def blabla(ind):
    fitness = np.random.random(1)
    features = np.random.random(3)
    return fitness, features



def sphere(ind):
    """Sphere function defined as:
    $$ f(x) = \sum_{i=1}^{n} x_i^{2} $$
    with a search domain of $-5.12 < x_i < 5.12, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(0, ..., 0) = 0.
    It is continuous, convex and unimodal.
    """
    return sum((x ** 2. for x in ind)),
illumination_sphere = algorithms.partial(_illuminate_artificial_landscape, func=sphere) # type: ignore
class SphereBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-5.12, 5.12)
        super().__init__(fn=algorithms.partial(illumination_sphere, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")

def weighted_sphere(ind):
    """Weighted sphere function defined as:
    $$ f(x) = \sum_{i=1}^{n} i \times x_i^{2} $$
    with a search domain of $-5.12 < x_i < 5.12, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(0, ..., 0) = 0.
    It is continuous, convex and unimodal.
    """
    return sum((i * (ind[i] ** 2.) for i in range(len(ind)))),
illumination_weighted_sphere = algorithms.partial(_illuminate_artificial_landscape, func=weighted_sphere) # type: ignore
class WeightedSphereBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-5.12, 5.12)
        super().__init__(fn=algorithms.partial(illumination_weighted_sphere, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def rotated_hyper_ellipsoid(ind):
    """Rotated hyper-ellipsoid function defined as:
    $$ f(x) = \sum_{i=1}^{n} \sum_{j=1}^{i} x_j^{2} $$
    with a search domain of $-65.536 < x_i < 65.536, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(0, ..., 0) = 0.
    It is continuous, convex and unimodal.
    """
    return sum(( sum((ind[j] ** 2. for j in range(i))) for i in range(len(ind)) )),
illumination_rotated_hyper_ellipsoid = algorithms.partial(_illuminate_artificial_landscape, func=rotated_hyper_ellipsoid) # type: ignore
class RotatedHyperEllipsoidBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-65.536, 65.536)
        super().__init__(fn=algorithms.partial(illumination_rotated_hyper_ellipsoid, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def rosenbrock(ind):
    """Rosenbrock function defined as:
    $$ f(x) = \sum_{i=1}^{n-1} 100 \times (x_{i+1} - x_i^{2})^{2} + (1 - x_{i})^{2} $$
    with a search domain of $-2.048 < x_i < 2.048, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(1, ..., 1) = 0.
    """
    return sum((100. * (ind[i+1]-ind[i]**2.)**2. + (1-ind[i])**2. for i in range(len(ind)-1))),
illumination_rosenbrock = algorithms.partial(_illuminate_artificial_landscape, func=rosenbrock) # type: ignore
class RosenbrockBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-2.048, 2.048)
        super().__init__(fn=algorithms.partial(illumination_rosenbrock, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def rastrigin_normalised(ind, a=10.):
    """Rastrigin test function, with inputs and outputs scaled to be between 0. and 1."""
    n = float(len(ind))
    ind_n = [-5.12 + x * (5.12 * 2.) for x in ind]
    return (a * n + sum(x * x - a * cos(2. * pi * x) for x in ind_n)) / (a * n + n * (5.12 * 5.12 + a)),
illumination_rastrigin_normalised = algorithms.partial(_illuminate_artificial_landscape, func=rastrigin_normalised) # type: ignore
class NormalisedRastriginBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (0., 1.)
        super().__init__(fn=algorithms.partial(illumination_rastrigin_normalised, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., 1.),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def rastrigin(ind, a=10.):
    """Rastrigin test function."""
    n = float(len(ind))
    return a * n + sum(x * x - a * cos(2. * pi * x) for x in ind),
illumination_rastrigin = algorithms.partial(_illuminate_artificial_landscape, func=rastrigin) # type: ignore
class RastriginBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-5.12, 5.12)
        super().__init__(fn=algorithms.partial(illumination_rastrigin, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def schwefel(ind):
    """Schwefel function defined as:
    $$ f(x) = 418.9829n - \sum_{i=1}^{n} x_i \sin(\sqrt(\norm(x_i))) $
    with a search domain of $-500 < x_i < 500, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(420.9687, ..., 420.9687) = 0.
    """
    return 418.9829 * float(len(ind)) - sum((x * math.sin(math.sqrt(abs(x))) for x in ind)),
illumination_schwefel = algorithms.partial(_illuminate_artificial_landscape, func=schwefel) # type: ignore
class SchwefelBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-500., 500.)
        super().__init__(fn=algorithms.partial(illumination_schwefel, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def small_schwefel(ind):
    """Schwefel function defined as:
    $$ f(x) = 418.9829n - \sum_{i=1}^{n} x_i \sin(\sqrt(\norm(x_i))) $
    with a search domain of $-100 < x_i < 100, 1 \leq i \leq n$.
    """
    return 418.9829 * float(len(ind)) - sum((x * math.sin(math.sqrt(abs(x))) for x in ind)),
illumination_small_schwefel = algorithms.partial(_illuminate_artificial_landscape, func=small_schwefel) # type: ignore
class SmallSchwefelBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        #ind_domain = (-50., 50.)
        ind_domain = (0., 500.)
        super().__init__(fn=algorithms.partial(illumination_small_schwefel, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def griewangk(ind):
    """Griewangk function defined as:
    $$ f(x) = 1 + \sum_{i=1}^{n} \frac{x_i^2}{4000} - \prod_{i=1}^n \cos \frac{x_i}{\sqrt{i}} $$
    with a search domain of $-600 < x_i < 600, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(0, ..., 0) = 0.
    """
    return 1. + sum(( (x ** 2.) / 4000. for x in ind)) - np.prod(list((math.cos(ind[i] / math.sqrt(i+1.)) for i in range(len(ind))))),
illumination_griewangk = algorithms.partial(_illuminate_artificial_landscape, func=griewangk) # type: ignore
class GriewangkBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-600., 600.)
        super().__init__(fn=algorithms.partial(illumination_griewangk, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def sum_of_powers(ind):
    """Sum Of Powers function defined as:
    $$ f(x) = \sum_{i=1}^{n} \norm{x_i}^{(i+1)} $$
    with a search domain of $-1 < x_i < 1, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(0, ..., 0) = 0.
    """
    return sum((abs(ind[i]) ** (i+1.) for i in range(len(ind)))),
illumination_sum_of_powers = algorithms.partial(_illuminate_artificial_landscape, func=sum_of_powers) # type: ignore
class SumOfPowersBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-1., 1.)
        super().__init__(fn=algorithms.partial(illumination_sum_of_powers, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def ackley(ind, a=20., b=0.2, c=2.*math.pi):
    """Ackley function defined as:
    $$ f(x) = -a \exp^{-b \sqrt{1/n \sum_{i=1}^n x_i^2} - \exp^{1/n \sum_{i=1}^n \cos{c x_i}} + a + \exp^{1} $$
    with $a=20, b=0.2, c=2\pi$ and with a search domain of $-32.768 < x_i < 32.768, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(0, ..., 0) = 0.
    """
    return - a * math.exp(-b * math.sqrt(1./len(ind) * sum((x**2. for x in ind)) )) - math.exp(1./len(ind) * sum((math.cos(c*ind[i]) for i in range(len(ind)) )) ) + a + math.exp(1.),
illumination_ackley = algorithms.partial(_illuminate_artificial_landscape, func=ackley) # type: ignore
class AckleyBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-32.768, 32.768)
        super().__init__(fn=algorithms.partial(illumination_ackley, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def styblinski_tang(ind):
    """Styblinski-Tang function defined as:
    $$ f(x) = 1/2 \sum_{i=1}^{n} x_i^4 - 16 x_i^2 + 5 x_i $$
    with a search domain of $-5 < x_i < 5, 1 \leq i \leq n$.
    """
    return sum(((x ** 4.) - 16. * (x ** 2.) + 5. * x for x in ind)) / 2.,
illumination_styblinski_tang = algorithms.partial(_illuminate_artificial_landscape, func=styblinski_tang) # type: ignore
class StyblinskiTangBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-5, 5)
        super().__init__(fn=algorithms.partial(illumination_styblinski_tang, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((-math.inf, math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def levy(ind):
    """Levy function defined as:
    $$ f(x) = \sin(\pi w_1)^2 + \sum_{i=1}^{n-1}(w_i-1)^2 (1+10 \sin(\pi w_i+1)^2) + (w_n-1)^2 (1+\sin(2\pi w_d)^2)$$
    where $w_i = 1+ \frac{x_i-1}{4}$
    with a search domain of $-10 < x_i < 10, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(1, ..., 1) = 0.
    """
    w = list((1. + (x-1.)/4. for x in ind))
    return (math.sin(math.pi * w[0]))**2. + \
        sum(((w[i]-1.)**2. * (1.+10.*(math.sin(math.pi*w[i]+1.)**2.)) for i in range(len(w)))) + \
        (w[-1]-1.)**2. * (1.+(math.sin(2.*math.pi*w[-1]))**2.),
illumination_levy = algorithms.partial(_illuminate_artificial_landscape, func=levy) # type: ignore
class LevyBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-10., 10.)
        super().__init__(fn=algorithms.partial(illumination_levy, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def perm0db(ind, beta=10.):
    """Perm function 0, D, BETA defined as:
    $$ f(x) = \sum_{i=1}^n (\sum_{j=1}^n (j+\beta) (x_j^i - \frac{1}{j^i}) )^2$$
    with a search domain of $-n < x_i < n, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(1, 2, ..., n) = 0.
    """
    return sum(( \
        ( sum(( (j+1.+beta) * ((ind[j])**float(i+1) - 1./(float(j+1)**float(i+1))) for j in range(len(ind)) )) )**2. \
        for i in range(len(ind)) )),
illumination_perm0db = algorithms.partial(_illuminate_artificial_landscape, func=perm0db) # type: ignore
class Perm0dbBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (float(-nb_features), float(nb_features))
        super().__init__(fn=algorithms.partial(illumination_perm0db, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def permdb(ind, beta=0.5):
    """Perm function D, BETA defined as:
    $$ f(x) = \sum_{i=1}^d (\sum_{j=1}^n (j^i+\beta) ((\frac{x_j}{j})^i - 1) )^2$$
    with a search domain of $-n < x_i < n, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(1, 1/2, ..., 1/n) = 0.
    """
    return sum(( \
        ( sum(( ((j+1.)**(i+1.)+beta) * ((ind[j]/(j+1.))**(i+1.) - 1.) for j in range(len(ind)) )) )**2. \
        for i in range(len(ind)) )),
illumination_permdb = algorithms.partial(_illuminate_artificial_landscape, func=permdb) # type: ignore
class PermdbBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (float(-nb_features), float(nb_features))
        super().__init__(fn=algorithms.partial(illumination_permdb, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def trid(ind):
    """Trid function defined as:
    $$ f(x) = \sum_{i=1}^n (x_i - 1)^2 - \sum_{i=2}^n x_i x_{i-1}$$
    with a search domain of $-n^2 < x_i < n^2, 1 \leq i \leq n$.
    The global minimum is $f(x^{*}) = -n(n+4)(n-1)/6$ at $x_i = i(n+1-i)$ for all $i=1,2,...,n$.
    """
    return sum(( (x-1)**2. for x in ind)) - sum(( ind[i] * ind[i-1] for i in range(1, len(ind)) )),
illumination_trid = algorithms.partial(_illuminate_artificial_landscape, func=trid) # type: ignore
class TridBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-float(nb_features)**2., float(nb_features)**2.)
        super().__init__(fn=algorithms.partial(illumination_trid, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def zakharov(ind):
    """Zakharov function defined as:
    $$ f(x) = \sum_{i=1}^n x_i^2 + (\sum_{i=1}^n 0.5 i x_i)^2 + (\sum_{i=1}^n 0.5 i x_i)^4$$
    with a search domain of $-5 < x_i < 10, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(0, ..., 0) = 0.
    """
    return  sum((x**2. for x in ind)) + \
            sum(( 0.5 * float(i) * ind[i] for i in range(len(ind)) ))**2. + \
            sum(( 0.5 * float(i) * ind[i] for i in range(len(ind)) ))**4.,
illumination_zakharov = algorithms.partial(_illuminate_artificial_landscape, func=zakharov) # type: ignore
class ZakharovBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-5., 10.)
        super().__init__(fn=algorithms.partial(illumination_zakharov, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def dixon_price(ind):
    """Dixon Price function defined as:
    $$ f(x) = (x_0 - 1)^2 + \sum_{i=2}^n i (2 x_i^2 - x_{i-1})^2$$
    with a search domain of $-10 < x_i < 10, 1 \leq i \leq n$.
    The global minimum is $f(x^{*}) = 0$ at $x_i = 2^{-(2^i-2)/(2^i)}$ for all $i=1,2,...,n$.
    """
    return  (ind[0] - 1.)**2. + \
            sum(( float(i) * (2.*(ind[i])**2. - ind[i-1])**2. for i in range(len(ind)) )),
illumination_dixon_price = algorithms.partial(_illuminate_artificial_landscape, func=dixon_price) # type: ignore
class DixonPriceBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-10., 10.)
        super().__init__(fn=algorithms.partial(illumination_dixon_price, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def powell(ind):
    """Powell function defined as:
    $$ f(x) = \sum_{i=1}^{n/4} ( (x_{4i-3}+10x_{4i-2})^2. + 5(x_{4i-1}-x_{4i})^2 + (x_{4i-2}-2x_{4i-1})^4 + 10(x_{4i-3}-x{4i})^4 )$$
    with a search domain of $-4 < x_i < 5, 1 \leq i \leq n$.
    The global minimum is at $f(x_1, ..., x_n) = f(0, ..., 0) = 0.
    """
    return  sum(( (ind[4*i-3] + 10.*ind[4*i-2])**2. for i in range(len(ind)//4) )) + \
            sum(( 5.* ((ind[4*i-1] - ind[4*i])**2.) for i in range(len(ind)//4) )) + \
            sum(( (ind[4*i-2] - 2.*ind[4*i-1])**4. for i in range(len(ind)//4) )) + \
            sum(( 10.* ((ind[4*i-3] - ind[4*i])**4.) for i in range(len(ind)//4) )),
illumination_powell = algorithms.partial(_illuminate_artificial_landscape, func=powell) # type: ignore
class PowellBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-4., 5.)
        super().__init__(fn=algorithms.partial(illumination_powell, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def michalewicz(ind, m=10.):
    """Michalewicz function defined as:
    $$ f(x) = - \sum_{i=1}^{n} \sin(x_i) \sin( \frac{i x_i^2}{\pi} )^(2*m)$$
    with a search domain of $0 < x_i < \pi, 1 \leq i \leq n$.
    """
    return - sum(( math.sin(ind[i]) * (math.sin(((i+1.)*(ind[i] **2.))/math.pi))**(2.*m) for i in range(len(ind)) )),
illumination_michalewicz = algorithms.partial(_illuminate_artificial_landscape, func=michalewicz) # type: ignore
class MichalewiczBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (0., math.pi)
        super().__init__(fn=algorithms.partial(illumination_michalewicz, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((-math.inf, 0.),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def wavy(ind, k=10.):
    """Wavy function defined as:
    $$ f(x) = 1 - \frac{1}{n} \sum_{i=1}^{n} \cos(kx_i)e^{-\frac{x_i^2}{2}}$$
    with a search domain of $-\pi < x_i < \pi, 1 \leq i \leq n$.
    """
    return 1. - sum(( math.cos(k * ind[i]) * math.exp(-(ind[i]*ind[i])/2.) for i in range(len(ind)))) / float(len(ind)),
illumination_wavy = algorithms.partial(_illuminate_artificial_landscape, func=wavy) # type: ignore
class WavyBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-math.pi, math.pi)
        super().__init__(fn=algorithms.partial(illumination_wavy, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., math.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def trigonometric02(ind, k=10.):
    """Trigonometric02 function defined as:
    $$ f(x) = 1 - \frac{1}{n} \sum_{i=1}^{n} \cos(kx_i)e^{-\frac{x_i^2}{2}}$$
    with a search domain of $-\pi < x_i < \pi, 1 \leq i \leq n$.
    """
    return 1. + sum((8.*math.sin(7. * (ind[i] - 0.9)**2.)**2. + 6.*math.sin(14.*(ind[i]-0.9)**2.)**2. + (ind[i] - 0.9)**2. for i in range(len(ind)))),
illumination_trigonometric02 = algorithms.partial(_illuminate_artificial_landscape, func=trigonometric02) # type: ignore
class Trigonometric02Benchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-500., 500.)
        super().__init__(fn=algorithms.partial(illumination_trigonometric02, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((1., np.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def qing(ind, k=10.):
    """Qing function defined as:
    $$ f(x) = \sum_{i=1}^{n} (x_i^2 - i)^2$$
    with a search domain of $-500 < x_i < 500, 1 \leq i \leq n$.
    """
    return sum(( (ind[i]**2. - i)**2. for i in range(len(ind)))),
illumination_qing = algorithms.partial(_illuminate_artificial_landscape, func=qing) # type: ignore
class QingBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-500., 500.)
        super().__init__(fn=algorithms.partial(illumination_qing, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., np.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")


def small_qing(ind, k=10.):
    """Qing function defined as:
    $$ f(x) = \sum_{i=1}^{n} (x_i^2 - i)^2$$
    with a search domain of $-4 < x_i < 4, 1 \leq i \leq n$.
    """
    return sum(( (ind[i]**2. - i)**2. for i in range(len(ind)))),
illumination_small_qing = algorithms.partial(_illuminate_artificial_landscape, func=small_qing) # type: ignore
class SmallQingBenchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-4., 4.)
        super().__init__(fn=algorithms.partial(illumination_small_qing, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((0., np.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")




def deb01(ind):
    """Deb01 function defined as:
    $$ f(x) = - \frac{1}{n} \sum_{i=1}^{n} \sin(5 \pi x_i)^6 $$
    with a search domain of $-1 < x_i < 1, 1 \leq i \leq n$.
    """
    return -sum(( math.sin(5. * math.pi * ind[i])**6. for i in range(len(ind)))) / float(len(ind)),
illumination_deb01 = algorithms.partial(_illuminate_artificial_landscape, func=deb01) # type: ignore
class Deb01Benchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-1., 1.)
        super().__init__(fn=algorithms.partial(illumination_deb01, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((-1., 1.),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")



def shubert04(ind):
    """Shubert04 function defined as:
    $$ f(x) = \sum_{i=1}^n \sum_{j=1}^5 j cos((j+1)x_i) + j $$
    with a search domain of $-10 < x_i < 10, 1 \leq i \leq n$.
    """
    return sum(( sum((-j * math.cos((j+1) * ind[i] + j) for j in range(1, 6))) for i in range(len(ind)))),
illumination_shubert04 = algorithms.partial(_illuminate_artificial_landscape, func=shubert04) # type: ignore
class Shubert04Benchmark(Benchmark):
    def __init__(self, nb_features: int = 2):
        self.nb_features = nb_features
        ind_domain = (-10., 10.)
        super().__init__(fn=algorithms.partial(illumination_shubert04, nb_features=nb_features),
                ind_domain = ind_domain,
                fitness_domain = ((-29.016015, np.inf),), features_domain = (ind_domain,) * nb_features,
                default_task = "minimisation")





# MODELINE "{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""Basic operators and tools used to design QD or evoluationary algorithms."""

import numpy as np
import random
from typing import Sequence, Callable, Tuple
from itertools import repeat
import copy

from qdpy.base import *
from qdpy import containers


########### SELECTION ########### {{{1

def sel_random(collection: Sequence[Any]) -> Sequence[Any]:
    """Select and return one individual at random among `collection`.

    Parameters
    ----------
    :param collection: Sequence[Any]
        The individuals to select from.
    """
    return random.choice(collection)


def non_trivial_sel_random(container: containers.Container) -> Any:
    """Select and return one individual at random among `container`.
    Ignore individuals with a trivial fitness (i.e. equal to the minimum fitness value),
    except if there are only individuals with trivial fitness.

    Parameters
    ----------
    :param container: Container
        The individuals to select from.
    """
    fst_ind: Any = container[0]
    min_fitness: Any = copy.deepcopy(fst_ind.fitness)
    assert container.fitness_domain is not None and len(min_fitness.values) == len(container.fitness_domain), f"You must specify `fitness_domain` in Container, and use individuals with fitness values of the same length as `fitness_domain`."
    min_fitness.values = tuple([x[0] for x in container.fitness_domain])
    candidates: MutableSequence[Any] = [ind for ind in container if ind.fitness.dominates(min_fitness)]
    if len(candidates):
        return random.choice(candidates)
    else:
        return random.choice(container)


########### MUTATIONS ########### {{{1

def mut_gaussian(individual: MutableSequence[Any], mu: float, sigma: float, mut_pb: float) -> MutableSequence[Any]:
    """Return a gaussian mutation of mean `mu` and standard deviation `sigma`
    on selected items of `individual`. `mut_pb` is the probability for each
    item of `individual` to be mutated.
    Mutations are applied directly on `individual`, which is then returned.

    Parameters
    ----------
    :param individual
        The individual to mutate.
    :param mu: float
        The mean of the gaussian mutation.
    :param sigma: float
        The standard deviation of the gaussian mutation.
    :param mut_pb: float
        The probability for each item of `individual` to be mutated.
    """
    for i in range(len(individual)):
        if random.random() < mut_pb:
            individual[i] += random.gauss(mu, sigma)
    return individual



def mut_polynomial_bounded(individual: MutableSequence[Any], eta: float, low: float, up: float, mut_pb: float) -> MutableSequence[Any]:
    """Return a polynomial bounded mutation, as defined in the original NSGA-II paper by Deb et al.
    Mutations are applied directly on `individual`, which is then returned.
    Inspired from code from the DEAP library (https://github.com/DEAP/deap/blob/master/deap/tools/mutation.py).

    Parameters
    ----------
    :param individual
        The individual to mutate.
    :param eta: float
        Crowding degree of the mutation.
        A high ETA will produce mutants close to its parent,
        a small ETA will produce offspring with more differences.
    :param low: float
        Lower bound of the search domain.
    :param up: float
        Upper bound of the search domain.
    :param mut_pb: float
        The probability for each item of `individual` to be mutated.
    """
    for i in range(len(individual)):
        if random.random() < mut_pb:
            x = individual[i]
            delta_1 = (x - low) / (up - low)
            delta_2 = (up - x) / (up - low)
            rand = random.random()
            mut_pow = 1. / (eta + 1.)

            if rand < 0.5:
                xy = 1. - delta_1
                val = 2. * rand + (1. - 2. * rand) * xy**(eta + 1.)
                delta_q = val**mut_pow - 1.
            else:
                xy = 1. - delta_2
                val = 2. * (1. - rand) + 2. * (rand - 0.5) * xy**(eta + 1.)
                delta_q = 1. - val**mut_pow

            x += delta_q * (up - low)
            x = min(max(x, low), up)
            individual[i] = x
    return individual




########### COMBINED OPERATORS ########### {{{1

def sel_or_init(collection: Sequence[IndividualLike], base_ind: IndividualLike,
        sel_fn: Callable, sel_pb: float,
        init_fn: Callable, init_pb: float = 0., return_flag: bool = True):
    """Either select an individual from `collection` by using function `sel_pb`,
    or initialise a new individual by using function `init_pb`.
    If `collection` is empty, it will always initialise a new individual, not perform selection.

    Parameters
    ----------
    :param collection: Sequence[IndividualLike]
        The individuals to select from.
    :param base_ind: IndividualLike
        The base individual to initialise.
    :param sel_fn: Callable
        The selection function.
    :param sel_pb: float
        The probability of performing selection.
    :param init_fn: Callable
        The initialisation function.
    :param init_pb: float
        The probability of performing initialisation.
    :param return_flag: bool
        If set to True, the function will return a Tuple[IndividualLike, bool] with a first item corresponding
        to the selected or initialised individual, and the second item a flag set to True if the first item
        was selected, and set to False if it was initialised.
        If set to False, the function will return the selected or initialised IndividualLike.
    """
    def ret(res, f):
        return (res, f) if return_flag else res
    if len(collection) == 0:
        return ret(init_fn(base_ind), False)
    operation = np.random.choice(range(2), p=[sel_pb, init_pb])
    if operation == 0: # Selection
        return ret(sel_fn(collection), True)
    else: # Initialisation
        return ret(init_fn(base_ind), False)


def mut_or_cx(individuals: Union[IndividualLike, Sequence[IndividualLike]],
        mut_fn: Callable, cx_fn: Callable) -> Sequence[IndividualLike]:
    """Either perform a mutation (using function `mut_fn`) or a crossover (using function `cx_fn`)
    depending on the nature and length of `individuals`.
    If `individuals` is an IndividualLike or a Sequence of one IndividualLike, a mutation will be performed.
    If `individuals` is a Sequence of two IndividualLike, a crossover will be performed.

    Parameters
    ----------
    :param individuals: Union[IndividualLike, Sequence[IndividualLike]]
        The individual(s) to mutate or crossover.
    :param mut_fn: Callable
        The mutation function.
    :param cx_fn: Callable
        The crossover function.

    Return
    ------
    The resulting individual(s).
    """
    if isinstance(individuals, IndividualLike):
        return [mut_fn(individuals)]
    elif isinstance(individuals, Sequence) and len(individuals) == 1:
        return [mut_fn(individuals[0])]
    elif isinstance(individuals, Sequence) and len(individuals) > 1:
        return cx_fn(individuals)
    else:
        raise ValueError(f'`individuals` can be an Individual or a Sequence.')


# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""A collection of all algorithms based on Evolutionary algorithms."""

__all__ = ["Evolution", "RandomSearchMutPolyBounded", "MutPolyBounded", "RandomSearchMutGaussian", "MutGaussian", "CMAES"]

########### IMPORTS ########### {{{1

from timeit import default_timer as timer
from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, Mapping, MutableMapping, overload
import warnings
import numpy as np
import copy
from functools import partial
import random
from inspect import signature


from .base import *
from qdpy.utils import *
from qdpy.base import *
from qdpy.containers import *
from qdpy import tools


########### EVOLUTION CLASSES ########### {{{1

class Evolution(QDAlgorithm):
    """TODO"""
    _select_fn: Optional[Callable]
    _select_or_initialise_fn: Optional[Callable]
    _select_or_initialise_fn_nb_parameters: int
    _vary_fn: Callable[[IndividualLike], IndividualLike]
    _tell_fn: Optional[Callable]

    def __getstate__(self):
        odict = super().__getstate__()
        del odict['_select_fn']
        del odict['_select_or_initialise_fn']
        del odict['_vary_fn']
        del odict['_tell_fn']
        return odict

    def __init__(self, container: Container, budget: int,
            vary: Callable[[IndividualLike], IndividualLike],
            select: Optional[Callable[[Container], IndividualLike]] = None,
            select_or_initialise: Optional[Callable[[Container], Tuple[IndividualLike, bool]]] = None,
            tell: Optional[Callable] = None, **kwargs):
        super().__init__(container, budget, **kwargs)
        if select is not None and select_or_initialise is not None:
            raise ValueError("Only one of `select` and `select_or_initialise` can be provided.")
        self._select_fn = select
        self._select_or_initialise_fn = select_or_initialise
        self._vary_fn = vary # type: ignore
        self._tell_fn = tell # type: ignore
        # Find the number of parameters of self._select_or_initialise_fn
        if self._select_or_initialise_fn is not None:
            sig = signature(self._select_or_initialise_fn)
            self._select_or_initialise_fn_nb_parameters = len(sig.parameters)
            if self._select_or_initialise_fn_nb_parameters == 0:
                raise ValueError("`select_or_initialise` function must take at least one parameter.")

    def _internal_ask(self, base_ind: IndividualLike) -> IndividualLike:
        # Select the next individual
        perform_variation: bool = False
        if self._select_fn is not None:
            selected: IndividualLike = copy.deepcopy(self._select_fn(self.container))
            if not isinstance(selected, IndividualLike):
                if isinstance(selected, Sequence) and isinstance(selected[0], IndividualLike):
                    selected = selected[0]
                else:
                    raise RuntimeError("`select` function returned an unknown type of individual.")
            perform_variation = True
        elif self._select_or_initialise_fn is not None:
            if self._select_or_initialise_fn_nb_parameters == 1:
                selected, perform_variation = self._select_or_initialise_fn(self.container)
            elif self._select_or_initialise_fn_nb_parameters >= 2:
                selected, perform_variation = self._select_or_initialise_fn(self.container, base_ind)
            selected = copy.deepcopy(selected)
            if perform_variation and not isinstance(selected, IndividualLike):
                raise RuntimeError("`select_or_initialise` function returned an unknown type of individual.")
        else:
            raise RuntimeError("Either `select` or `select_or_initialise` must be provided.")

        # Vary the suggestion
        if perform_variation:
            varied = self._vary_fn(selected) # type: ignore
            if not isinstance(varied, IndividualLike):
                if is_iterable(varied) and isinstance(varied[0], IndividualLike):
                    varied = varied[0]
                elif is_iterable(varied):
                    pass # Try anyway
                else:
                    raise RuntimeError("`vary` function returned an unknown type of individual.")
                base_ind[:] = varied
            else:
                return varied
        else:
            if isinstance(selected, IndividualLike):
                return selected
            else:
                base_ind[:] = selected
        return base_ind

    def _internal_tell(self, individual: IndividualLike, added_to_container: bool) -> None:
        if self._tell_fn is not None:
            self._tell_fn(individual, added_to_container)




@registry.register
class RandomSearchMutPolyBounded(Evolution):
    """TODO"""
    ind_domain: DomainLike
    sel_pb: float
    init_pb: float
    mut_pb: float
    eta: float

    def __init__(self, container: Container, budget: int,
            dimension: int, ind_domain: DomainLike = (0., 1.),
            sel_pb: float = 0.5, init_pb: float = 0.5, mut_pb: float = 0.2, eta: float = 20.,
            **kwargs):
        self.ind_domain = ind_domain
        self.sel_pb = sel_pb
        self.init_pb = init_pb
        self.mut_pb = mut_pb
        self.eta = eta

        #init_fn = partial(np.random.uniform, ind_domain[0], ind_domain[1], dimension)
        init_fn = partial(lambda dim, base_ind: [random.uniform(ind_domain[0], ind_domain[1]) for _ in range(dim)], dimension)
        select_or_initialise = partial(tools.sel_or_init,
                sel_fn = tools.sel_random,
                sel_pb = sel_pb,
                init_fn = init_fn,
                init_pb = init_pb)
        vary = partial(tools.mut_polynomial_bounded, low=ind_domain[0], up=ind_domain[1], eta=eta, mut_pb=mut_pb)

        super().__init__(container, budget, dimension=dimension, # type: ignore
                select_or_initialise=select_or_initialise, vary=vary, **kwargs)


@registry.register
class MutPolyBounded(RandomSearchMutPolyBounded):
    """TODO"""
    def __init__(self, container: Container, budget: int,
            dimension: int, ind_domain: DomainLike = (0., 1.),
            mut_pb: float = 0.2, eta: float = 20., **kwargs):
        super().__init__(container, budget, dimension=dimension,
                ind_domain=ind_domain,
                sel_pb=1.0, init_pb=0.0,
                mut_pb=mut_pb, eta=eta, **kwargs)


@registry.register
class RandomSearchMutGaussian(Evolution):
    """TODO"""
    sel_pb: float
    init_pb: float
    mut_pb: float
    mu: float
    sigma: float

    def __init__(self, container: Container, budget: int,
            dimension: int, sel_pb: float = 0.5, init_pb: float = 0.5, mut_pb: float = 0.2,
            mu: float = 0., sigma: float = 1.0, **kwargs):
        self.sel_pb = sel_pb
        self.init_pb = init_pb
        self.mut_pb = mut_pb
        self.mu = mu
        self.sigma = sigma

        #init_fn = partial(np.random.normal, self.mu, self.sigma, dimension)
        init_fn = partial(lambda dim, base_ind: [random.normalvariate(self.mu, self.sigma) for _ in range(dim)], dimension)
        select_or_initialise = partial(tools.sel_or_init,
                sel_fn = tools.sel_random,
                sel_pb = sel_pb,
                init_fn = init_fn,
                init_pb = init_pb)
        vary = partial(tools.mut_gaussian, mu=mu, sigma=sigma, mut_pb=mut_pb)

        super().__init__(container, budget, dimension=dimension, # type: ignore
                select_or_initialise=select_or_initialise, vary=vary, **kwargs)


@registry.register
class MutGaussian(RandomSearchMutGaussian):
    """TODO"""
    def __init__(self, container: Container, budget: int,
            dimension: int, mut_pb: float = 0.2,
            mu: float = 0., sigma: float = 1.0, **kwargs):
        super().__init__(container, budget, dimension=dimension,
                sel_pb=1.0, init_pb=0.0,
                mut_pb=mut_pb, mu=mu, sigma=sigma, **kwargs)


try:
    import cma

    @registry.register
    class CMAES(QDAlgorithm):
        """TODO"""
        ind_domain: Optional[DomainLike]
        sigma0: float
        es: Any
        _pop_inds: MutableSequence[IndividualLike]
        _pop_fitness_vals: MutableSequence[Any]

        def __init__(self, container: Container, budget: int,
                dimension: int, ind_domain: Optional[DomainLike] = None,
                sigma0: float = 1.0,
                ignore_if_not_added_to_container: bool = False, **kwargs):
            super().__init__(container, budget, dimension=dimension, **kwargs)
            self.sigma0 = sigma0
            self.ind_domain = ind_domain
            self.ignore_if_not_added_to_container = ignore_if_not_added_to_container
            self._opts: MutableMapping[str, Any] = {}
            if ind_domain is not None:
                self._opts['bounds'] = list(ind_domain)
            if self._batch_size is not None:
                self._opts['popsize'] = self._batch_size
            self.es = cma.CMAEvolutionStrategy([0.] * dimension, sigma0, self._opts)
            self._pop_inds = []
            self._pop_fitness_vals = []
            if self._batch_size is None:
                self._batch_size = self.es.popsize

        def _internal_ask(self, base_ind: IndividualLike) -> IndividualLike:
            base_ind[:] = self.es.ask(1)[0]
            return base_ind

        def _internal_tell(self, individual: IndividualLike, added_to_container: bool) -> None:
            if self.ignore_if_not_added_to_container and not added_to_container:
                return
            self._pop_inds += [individual]
            self._pop_fitness_vals += [-1. * x for x in individual.fitness.values]
            if len(self._pop_inds) >= self.es.popsize:
                try:
                    self.es.tell(self._pop_inds, self._pop_fitness_vals)
                except RuntimeError:
                    pass
                else:
                    self._pop_inds.clear()
                    self._pop_fitness_vals.clear()


except ImportError:
    @registry.register # type: ignore
    class CMAES(QDAlgorithm):
        def __init__(self, *args, **kwargs):
            raise NotImplementedError("`CMAES` needs the 'cma' package to be installed and importable.")
        def _internal_ask(self, base_ind: IndividualLike) -> IndividualLike:
            raise NotImplementedError("`CMAES` needs the 'cma' package to be installed and importable.")




# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#!/usr/bin/env python3


from setuptools import setup, find_packages
import qdpy

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

extras_require = {
    'all': [
        'deap>=1.2.2',
        'tqdm>=4.28.1',
        'colorama>=0.4.1',
        'tables>=3.4.4',
        'cma>=2.6.0',
    ],
    'deap': [
        'deap>=1.2.2',
    ]
}

packages = find_packages(exclude=['examples'])

setup(name='qdpy',
        long_description=qdpy.__doc__,
        version=qdpy.__version__,
        description='Quality-Diversity algorithms in Python',
        url='https://gitlab.com/leo.cazenille/qdpy',
        author='Leo Cazenille',
        author_email='leo.cazenille@gmail.com',
        license='LGPLv3',
        packages=packages,
        python_requires=">=3.6",
        install_requires=requirements,
        extras_require=extras_require,
        zip_safe=False,
        classifiers = [
            "Intended Audience :: Science/Research",
            "Intended Audience :: Education",
            "Intended Audience :: Other Audience",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Mathematics",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Visualization",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Development Status :: 2 - Pre-Alpha",
            "Environment :: Console",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        ],
        keywords=["optimisation", "optimization", "illumination", "Quality-Diversity", "Landscape exploration", "MAP-Elites", "NSLC", "CVT-MAP-Elites", "SAIL", "evolutionary algorithms", "random search"]
)

# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""TODO"""

import numpy as np
#from scipy.spatial.distance import euclidean
#from itertools import starmap
from typing import Sequence, Callable, Tuple

from qdpy.base import *


########### METRICS ########### {{{1

#@jit(nopython=True)
def features_distances(individual: IndividualLike, container: Sequence, dist: Union[str, Callable] = "euclidean") -> Sequence:
    distances = np.zeros(len(container))
    ind_features = individual.features
    if isinstance(dist, str):
        if dist == "euclidean":
            for i, other in enumerate(container):
                other_features = other.features
                for j in range(len(ind_features)):
                    distances[i] += pow(ind_features[j] - other_features[j], 2.)
            distances = np.power(distances, 1./2.)
        else:
            raise ValueError(f"Unknown `dist` type: '{dist}'.")
    else:
        for i, ind in enumerate(container):
            distances[i] = dist(ind_features, ind.features)
    return distances

def novelty(individual: IndividualLike, container: Sequence, k: int = 1, dist: Union[str, Callable] = "euclidean", ignore_first: bool = False, default_novelty: float = 0.1) -> float:
    """Returns the novelty score of ``individual`` in ``container``.
    Novelty is defined as the average distance to the ``k``-nearest neighbours of ``individual``. If ``container`` is empty, return ``default_novelty``."""
    if len(container) == 0:
        return default_novelty
    n_k = min(len(container), k)
    distances: Sequence = features_distances(individual, container, dist)
    if ignore_first:
        nearest_neighbours_dists: Sequence = sorted(distances)[1:n_k+1]
    else:
        nearest_neighbours_dists = sorted(distances)[:n_k]
    return np.mean(nearest_neighbours_dists)


def novelty_nn(individual: IndividualLike, container: Sequence, k: int = 1, nn_size: int = 1, dist: Union[str, Callable] = "euclidean", ignore_first: bool = False, default_novelty: float = 0.1) -> Tuple[float, Sequence]:
    """Returns the novelty score of ``individual`` in ``container`` and the indexes of its ``nn_size`` nearest neighbours.
    Novelty is defined as the average distance to the ``k``-nearest neighbours of ``individual``.  If ``container`` is empty, return ``default_novelty``."""
    if len(container) == 0:
        return default_novelty, []
    n_k = min(len(container), k)
    n_nn_size = min(len(container), nn_size)
    distances: Sequence = features_distances(individual, container, dist)
    idx_container = list(range(len(container)))
    if ignore_first:
        nearest_neighbours_dists: Sequence = sorted(distances)[1:n_k+1]
        nn: Sequence = sorted(zip(distances, idx_container))[1:n_nn_size+1]
    else:
        nearest_neighbours_dists = sorted(distances)[:n_k]
        nn = sorted(zip(distances, idx_container))[:n_nn_size]
    novelty: float = np.mean(nearest_neighbours_dists)
    nearest_neighbours_idx: Sequence
    _, nearest_neighbours_idx = tuple(zip(*nn))
    return novelty, nearest_neighbours_idx


def novelty_local_competition(individual: IndividualLike, container: Sequence, k: int = 1, dist: Union[str, Callable] = "euclidean", ignore_first: bool = False, default_novelty: float = 0.1, default_local_competition: float = 1.0) -> Tuple[float, float]:
    """Returns the novelty and normalised local competition scores of ``individual`` in ``container``.
    Novelty is defined as the average distance to the ``k``-nearest neighbours of ``individual``.
    Local competition is defined as the number of ``k``-nearest neighbours of ``individual`` that are outperformed by ``individual``. This value is normalised by ``k`` to be in domain [0., 1.].
    If ``container`` is empty, return ``default_novelty`` and ``default_local_competition``."""
    if len(container) == 0:
        return default_novelty, default_local_competition
    distances: Sequence = features_distances(individual, container, dist)
    nearest_neighbours_dists: Sequence
    nearest_neighbours: Sequence
    if ignore_first:
        nn: Sequence = sorted(zip(distances, container))[1:k+1]
    else:
        nn = sorted(zip(distances, container))[:k]
    nearest_neighbours_dists, nearest_neighbours = tuple(zip(*nn))
    novelty: float = np.mean(nearest_neighbours_dists)
    local_competition: float = sum((individual.fitness.dominates(ind.fitness) for ind in nearest_neighbours)) / float(k)
    return novelty, local_competition




# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""Package `qdpy` implements recent Quality-Diversity algorithms: Map-Elites, CVT-Map-Elites, NSLC, SAIL, etc. 
QD algorithms can be accessed directly, but `qdpy` also includes building blocks that can be easily assembled together to build your own QD algorithms. It can be used with parallelism mechanisms and in distributed environments.

This package requires Python 3.6+.

`qdpy` includes the following features:
 * Generic support for diverse Containers: Grids, Novelty-Archives, Populations, etc
 * Optimisation algorithms for QD: random search methods, quasi-random methods, evolutionary algorithms
 * Support for multi-objective optimisation methods
 * Possible to use optimisation methods not designed for QD, such as [CMA-ES](https://arxiv.org/pdf/1604.00772.pdf)
 * Parallelisation of evaluations, using parallelism libraries, such as multiprocessing, concurrent.futures or [SCOOP](https://github.com/soravux/scoop)
 * Easy integration with the popular [DEAP](https://github.com/DEAP/deap) evolutionary computation framework 

Install
=======
`qdpy` requires Python 3.6+. It can be installed with:
    pip3 install qdpy

`qdpy` includes optional features that need extra packages to be installed:
 * `cma` for CMA-ES support
 * `deap` to integrate with the DEAP library
 * `tables` to output results files in the HDF5 format
 * `tqdm` to display a progress bar showing optimisation progress
 * `colorama` to add colours to pretty-printed outputs

You can install `qdpy` and all of these optional dependencies with:
    pip3 install qdpy[all]

The latest version can be installed from the GitLab repository:
    pip3 install git+https://gitlab.com/leo.cazenille/qdpy.git@master

Example
=======
From a python shell::

    from qdpy import algorithms, containers, benchmarks, plots
    
    # Create container and algorithm. Here we use MAP-Elites, by illuminating a Grid container by evolution.
    grid = containers.Grid(shape=(64,64), max_items_per_bin=1, fitness_domain=((0., 1.),), features_domain=((0., 1.), (0., 1.)))
    algo = algorithms.RandomSearchMutPolyBounded(grid, budget=60000, batch_size=500,
            dimension=3, optimisation_task="maximisation")
    
    # Create a logger to pretty-print everything and generate output data files
    logger = algorithms.AlgorithmLogger(algo)
    
    # Define evaluation function
    eval_fn = algorithms.partial(benchmarks.illumination_rastrigin_normalised,
            nb_features = len(grid.shape))
    
    # Run illumination process !
    best = algo.optimise(eval_fn)
    
    # Print results info
    print(algo.summary())
    
    # Plot the results
    plots.default_plots_grid(logger)
    
    print("All results are available in the '%s' pickle file." % logger.final_filename)


Usage, Documentation
====================
Please to go the GitLab repository main page (https://gitlab.com/leo.cazenille/qdpy) and the documentation main page (https://leo.cazenille.gitlab.io/qdpy/).


:Author: Leo Cazenille, 2018-*

:License: LGPLv3, see LICENSE file.

"""


__author__ = "qdpy Team"
__license__ = "LGPLv3"
__version__ = "0.0.9.8"
__revision__ = "0.0.9.8"

# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker



#    This file is part of qdpy.
#
#    qdpy is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    qdpy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with qdpy. If not, see <http://www.gnu.org/licenses/>.

"""TODO""" # TODO

__all__ = ["RandomUniform", "Sobol"]

from typing import Optional, Tuple, List, Iterable, Iterator, Any, TypeVar, Generic, Union, Sequence, MutableSet, MutableSequence, Type, Callable, Generator, Mapping, MutableMapping, overload
import numpy as np
import random


from .base import *
from qdpy.utils import *
from qdpy.base import *
from qdpy.containers import *
from qdpy import tools

import qdpy.hdsobol as hdsobol


@registry.register
class RandomUniform(QDAlgorithm):
    """TODO"""
    ind_domain: DomainLike
    dimension: int

    def __init__(self, container: Container, budget: int, ind_domain: DomainLike, **kwargs):
        super().__init__(container, budget, **kwargs)
        self.ind_domain = ind_domain
        if self.dimension is None:
            raise ValueError("`dimension` must be provided.")

    def _internal_ask(self, base_ind: IndividualLike) -> IndividualLike:
        #base_ind[:] = np.random.uniform(self.ind_domain[0], self.ind_domain[1], size=self.dimension)
        base_ind[:] = [random.uniform(self.ind_domain[0], self.ind_domain[1]) for _ in range(self.dimension)]
        return base_ind



@registry.register
class Sobol(RandomUniform):
    """TODO"""

    def __init__(self, container: Container, budget: int, ind_domain: DomainLike, **kwargs):
        super().__init__(container, budget, ind_domain=ind_domain, **kwargs)
        self.sobol_vect = np.array(hdsobol.gen_sobol_vectors(self.budget + 1, self.dimension), dtype=float)
        self.sobol_vect = self.sobol_vect * (ind_domain[1] - ind_domain[0]) + ind_domain[0]

    def _internal_ask(self, base_ind: IndividualLike) -> IndividualLike:
        base_ind[:] = self.sobol_vect[self._nb_suggestions]
        return base_ind



# MODELINE	"{{{1
# vim:expandtab:softtabstop=4:shiftwidth=4:fileencoding=utf-8
# vim:foldmethod=marker
